#!/usr/bin/python3

import re
from scipy.stats import rv_discrete
import sys
import numpy as np

# we should define these somewhere that is associated with the
# evaluation_functino. Maybe member variables that are initialized in
# the print_dot function ... but that would make evaluation depend on
# the output format, which is not very elegant.
# technically we should use the apta after merging, and
MEAN_REGEX = '(?P<state>\d+) \[shape=(doublecircle|circle|ellipse) label=\"\[(?P<mean>\d+)\].*\"\];'
SYMLST_REGEX = "((?P<sym>-?\d+) \[(?P<occ>\d+):\d+\])+"
TRAN_REGEX = "^(?P<sst>.+) -> (?P<dst>.+) \[label=\"(?P<slst>.+)\"[ style=dotted]*\];$"

def load_means(content):
    means = []

    for line in content.split("\n"):
      matcher = re.match(MEAN_REGEX, line.strip())
#      print(line.strip())
#      print(MEAN_REGEX)
      if matcher is None: continue
      #for match in matcher:
      match = matcher
      if match.group("mean") == "0": continue
      means.append("{} -> {} [label=\" {} [{}:0]\"];".format(match.group("state"), match.group("state"),
                                                              "-1", match.group("mean")))
    return means

def load_model_from_file(dot_path, normalize=False):
    with open(dot_path, "r") as f:
        return load_model(f.read(), normalize)

# Read and extract automaton structure from dot file
# returns: dictionary dict[currect_state][symbol_read] = next_state
def load_model(dot_string, normalize=False, with_means=False):
    model = {}
    if with_means:
      means = "\n".join(load_means(dot_string))
      dot_string = means + dot_string

    for line in dot_string.split("\n"):
        matcher = re.match(TRAN_REGEX, line.strip())
        if matcher is not None:
            sstate = matcher.group("sst")
            dstate = matcher.group("dst")
            symlist = matcher.group("slst")
            if sstate not in model:
                model[sstate] = {}
            for smatcher in re.finditer(SYMLST_REGEX, symlist):
                symbol = int(smatcher.group("sym"))
                occurrences = int(smatcher.group("occ"))
                if symbol not in model[sstate]:
                    model[sstate][symbol] = (dstate, occurrences)
                else:
                    gamma = 0

    # normalizing occurrence counts to probabilities
    if normalize:
        for state in model:
            ssum = sum(map(lambda x: x[1], model[state].values()))
            for symbol in model[state]:
                dstate, occurrences = model[state][symbol]
                model[state][symbol] = (dstate, occurrences / float(ssum))
    return model



# given model, make predictions for all lines in test-path and
# write result to pred-path
def get_word_acceptance(model, test_array):
    REJECT = 0
    ACCEPT = 1

    accept = 0
    waccept = 0
    reject = 0
    okwindows = 0
    wneg = 0
    pred_array = []

    for line in test_array:
        state = "0"

        symbol_sequence=[]

        #for actual_symbol in sequence:
        for i in range(2, len(line.strip().split(" "))):
            symbol_sequence.append(int(line.strip().split(" ")[i].split(",")[0]))

        #print symbol_sequence

        # loop over word
        rejected = False
        for i in range(len(symbol_sequence)):
            actual_symbol = symbol_sequence[i]

            if state not in model:
                pred_array.append(REJECT)
                rejected = True
                break
            else:
                try:
                    state = model[state][actual_symbol][0]
                except KeyError:
                    rejected = True
                    pred_array.append(REJECT)
                    #print "in state %s read %s, no transition" % (state, actual_symbol)
                    #print model[state]
                    reject+=1
                    break

        if not rejected:
            pred_array.append(ACCEPT)
            #print "Accepted %s" % (accept+reject)
            accept+=1
            waccept+=1

        else:
            gamma = 0
            #print "Rejected"

        if ((accept+reject) % 5) == 0:
           if waccept > 3:
#             pred_array.append("windowok")
             #print "Window ok"
             okwindows+=1
           else:
             if waccept == 0:
               wneg+=1
           waccept = 0

    if accept == 0:
        #print "no acceptance"
        accept = 1
    if reject == 0:
        #print "no reject"
        reject = 1

     # this output is for the CNSM/LCN papers
#     if float(accept)/(accept+reject) > 0.75:
#         print("BOTNET")
#     else:
#         print("CLEAN")

     # this return should probably an array of (0/1, prob) for the
     # tst input argument instead of putting it into the output file
    print("Acc %s Rej %s, total %s, ratio %s, quota %s, Windows: %s of %s, %s neg" % (accept, reject, float(accept)/(accept+reject), float(accept)/reject, float(reject)/accept, okwindows, (accept+reject)/5, wneg))

    return pred_array


def predict(prefix, model):
  # traverse
  state = "0"
  for s in prefix.split(" "):
    if s == "": continue
    try:
      state = model[state][int(s)][0]
#      print(state)
    except:
#      print("hello")
      break
  if len(model[state].items()) == 0:
    return [(-1, (0, 1))]
  return sorted(model[state].items(), key=lambda x: x[1][1], reverse=True)

if __name__ == "__main__":

    # you can also invoke the script from the command line
    if len(sys.argv) > 1:
        dot = sys.argv[1]
        tst = sys.argv[2]
        of = sys.argv[3]
    else:
        print("usesage: ./name dotfile testfile outputfile")

    m = load_model_from_file(dot, normalize=True)

    # print m
    get_word_acceptance(m, tst, of)





import sys, os, pprint
from dfasat.estimators import DFASATEstimator, BaggingClassifier

if __name__ == '__main__':
    if len(sys.argv) < 2:
        sys.exit("Usage: " + __file__ + " train_file")

    train_file = sys.argv[1]

    file_prefix = os.path.splitext(os.path.basename(train_file))[0]

    with open(train_file) as f:
        train = f.read()

    train_data = train.strip().split("\n")
    train_data_x = []
    train_data_y = []
    for line in train_data[1:]:
        train_data_x.append([str(i) for i in line.split(" ")])
        train_data_y.append(1)

    pprint.pprint(train_data_x[:5])
    pprint.pprint(train_data_y[:5])

    bag = BaggingClassifier(estimator=DFASATEstimator, number=50, output_file=file_prefix+'5.5.05', random_seed=True, random_counts=None,#[5, 15, 25],
#                      hData='kl_data', hName='kldistance',
                      hData='alergia_data', hName='alergia',
#                      hData='overlap_data', hName='overlap_driven',
#                      hData='likelihood_data', hName='likelihoodratio',
                      symbol_count=5, # -y
                      state_count=5, # -t
#                      parameter=0.5, # -p
#                      extend=0, # -x
                      apta_bound=1, # -b
                      dfa_bound=2000, # -d
                      sinkson=0, # -i
                      method=1, # -m
                      tries=1 # -n
                      )


#    est = sklearn.ensemble.BaggingClassifier(base_est, 2)
    fit = bag.fit(train_data_x, train_data_y, subset=True)



"""
This is an estimator based on the DFASAT library developed at the University of Luxembourg
"""
import numpy as np
import scipy as sp
import numbers
from sklearn.base import BaseEstimator
import dfasat.lib.dfasat as dfasat
import dfasat.eval as e
import sys
from multiprocessing import Process, cpu_count
import random
import signal
signal.signal(signal.SIGINT, signal.SIG_DFL)

import os,sys,select
def more_data(elm):
    r, _, _ = select.select([elm],[],[],0)
    return bool(r)

def read_pipe(elm):
    out = ''
    while more_data(elm):
        out += " " + str(os.read(elm, 1024))
    return out

has_graphviz = True
try:
    from graphviz import Source
except:
    has_graphviz = False


class StateMachinePlot():
    """ A helper class implementing a _repr_svg_ method

    Parameters
    ----------
    svg : string
        The plot in SVG format
    """
    _svg = None

    def __init__(self, svg):
        self._svg = svg

    def _repr_svg_(self):
        return self._svg


num_cores = cpu_count()

class BaggingClassifier:
    estimator = None
    number = 10
    estimators = []
    file_prefix = "dfa"

    def __init__(self, estimator=estimator, number=10, output_file="dfa", random_seed=False, random_counts=None, **kwargs):
        self.estimator = estimator
        self.number = number
        self.file_prefix = output_file

        for _ in range(number):
            if random_seed:
                kwargs['seed'] = random.randint(0, 5000000)
            if random_counts is not None:
                kwargs['state_count'] = random.choice(random_counts)
                kwargs['symbol_count'] = random.choice(random_counts)

            self.estimators.append(estimator(**kwargs))

    def calc_model(self, i, estimator, X, y):
        estimator.fit(X, y)
        with open(self.file_prefix + '_' + str(i) + '.model.dot', 'w') as r:
            r.write(estimator.merger.dot_output)


    def fit(self, X, y, subset=True):
        processes = []
        for i, estimator in enumerate(self.estimators):
            if subset:
                print('randomly sampling')
                s_X=[]
                for x in X:
                    if random.randint(1, 5) != 1:
                        s_X.append(x)
                s_y = [1]*len(s_X)
            else:
                s_X = X
                s_y = y

            p = Process(target=self.calc_model, args=(i, estimator, s_X, s_y))
            processes.append(p)
            print('starting process '+str(i))
            p.start()
            if (i+1) % num_cores == 0:
                for p in processes:
                    p.join()
                    processes.remove(p)
        for p in processes:
            p.join()
            processes.remove(p)


    def predict(self, prefix):
        predictions = []
        for estimator in self.estimators:
            m = e.load_model(estimator.merger.dot_output, normalize=True)
            prediction = e.predict(prefix, m)
            predictions.append(prediction)

        sorted_predictions = {}
        for sample in predictions:
            for p in sample:
                if p[0] not in sorted_predictions:
                    sorted_predictions[p[0]] = p[1][1]
                else:
                    sorted_predictions[p[0]] += p[1][1]

        sorted_predictions = sorted(sorted_predictions.items(), key=lambda p: p[1], reverse=True)

        return sorted_predictions

class DFASATEstimator(BaseEstimator):
    """ An estimator that uses the DFASAT library

       import code
       code.interact(local=locals())

    """
    result_dot = None
    merger = None
    parameters = None
    output_file = "dfa"
    output_counter = 1

    def __init__(self, **kwargs):
        """hData="likelihood_data", hName="likelihoodratio",
            tries=1, sat_program="", dot_file="dfa", sinkson=1, seed=12345678,
            apta_bound=2000, dfa_bound=50, merge_sinks_p=1, merge_sinks_d=0,
            lower_bound=-1.0, offset=5, method=1, heuristic=1, extend=1, symbol_count=10,
            state_count=25, correction=1.0, parameters=0.5, extra_states=0,
            target_rejecting=0, symmetry=1, forcing=0"""

        self.parameters = dfasat.parameters()
        for key, value in kwargs.items():
            if key == 'output_file':
                self.output_file = value
                continue
            if not hasattr(self.parameters, key):
                raise Exception(key + " not a valid argument")
            setattr(self.parameters, key, value)
        """p.hData = hData
        p.hName = hName
        p.tries = tries
        p.sat_program = sat_program
        p.dot_file = dot_file
        p.sinkson = sinkson
        p.seed = seed
        p.apta_bound = apta_bound
        p.dfa_bound = dfa_bound
        p.merge_sinks_p = merge_sinks_p
        p.merge_sinks_d = merge_sinks_d
        p.lower_bound = lower_bound
        p.offset = offset
        p.method = method
        p.heuristic = heuristic
        p.extend = extend
        p.symbol_count = symbol_count
        p.state_count = state_count
        p.correction = correction
        p.parameters = parameters
        p.extra_states = extra_states"""


    def _verify_x(self, arr):
        if isinstance(arr, np.ndarray):
            x = arr
            arr = arr.flatten().tolist()
            if len(arr) == 0:
                raise ValueError("0 feature(s) (shape=" + str(x.shape) + ") while a minimum of " + str(x.shape[0]) + " is required.")
        if sp.sparse.issparse(arr):
            raise TypeError("sparse matrices not supported")
        for n in arr:
            if not isinstance(n, numbers.Number):
                pass
#                print(n)
#                raise TypeError("argument must be a string or a number")
#            if np.isinf(n) or np.isnan(n):
#                raise ValueError("argument cannot contain info or NaN")

    def _verify_y(self, arr):
        pass

    def _verify_args(self, x, y):
        if isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
            if x.shape[0] != y.shape[0]:
                raise ValueError("X and y must have the same number of samples")
        self._verify_x(x)
        self._verify_y(y)

    def _header(self, X):
        return str(len(X)) + " " + str(self._alphabet_size(X))

    def _alphabet_size(self, X):
        s = set()
        for row in X:
            for value in row[1:]:
                s.add(value.split("/")[0])
        print(s)
        return len(s)

    def _numpy_to_str(self, x):
        return [" ".join([str(int(symbol)) for symbol in sample]) for sample in x]

    def get_params(self, deep):
        params = {}
        for pName in dir(self.parameters):
            if not pName.startswith("__"):
                params[pName] = getattr(self.parameters, pName)
        return params

    def set_params(self, params):
        for pName, val in params.items():
            setattr(self.parameters, pName, val)


    def fit(self, x, y):
        # merge x and y into our format (array of strings)
        if type(x) == np.ndarray:
            x = self._numpy_to_str(x)
        self._verify_args(x, y)

        header = self._header(x)
        transformed_x = [" ".join([str(value) for value in row]) for row in x]
        for i in range(len(y)):
           transformed_x[i] = str(y[i]) + " " + transformed_x[i]

        transformed_x.insert(0, header)

        output_file = self.output_file + str(self.output_counter)
        self.output_counter += 1

        self.fit_(dfa_data=transformed_x, output_file=output_file)
        return self

    def fit_(self, dfa_file=None, dfa_data=None, output_file="dfa"):
        """A reference implementation of a fitting function

        Parameters
        ----------
        hData : string
            name of the evaluation data class
        hName : string
            name of the evaluation function
        tries : integer
            number of iterations
        sat_program : string
            path to the sat solver program
        dfa_file : string
            path to the dfa file
        dfa_data : list of strings
            dfa data as array of strings

        Returns
        -------
        self : object
            Returns self.
        """
        self.result_dot = None
        if dfa_file:
            print("dfa file set to " + dfa_file)
            self.parameters.dfa_file = dfa_file
        elif dfa_data:
            print("dfa_data set")
            self.parameters.dfa_data = dfasat.vector_less__std_scope_string__greater_()
            for s in dfa_data:
                self.parameters.dfa_data.append(s+"\n")
        else:
            raise TypeError('One of dfa_file or dfa_data has to be given')

        dfasat.init_with_params(self.parameters)
        a = dfasat.apta()
        l = getattr(dfasat, self.parameters.hName)()

        self.merger = dfasat.state_merger(l, a)
        if self.parameters.dfa_file:
            self.merger.read_apta(self.parameters.dfa_file)
        else:
            self.merger.read_apta(self.parameters.dfa_data)

        output_dot_file = output_file + ".dot"
        output_aut_file = output_file + ".aut"
        print('starting dfasat')
        sys.stdout.write('\b')
        pipe_out, pipe_in = os.pipe()
        stdout = os.dup(1)
        os.dup2(pipe_in, 1)
        dfasat.dfasat(self.merger, self.parameters.sat_program, output_dot_file, output_aut_file)
        print(read_pipe(pipe_out).replace('\\n', " | ").replace(" b\'", ""))
        os.dup2(stdout,1)
        print('dfasat done')
        self.result_dot = output_dot_file
        return self

    def predict(self, X):
#        self._verify_x(X)
        """ A reference implementation of a predicting function.

        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples.

        Returns
        -------
        y : array of shape = [n_samples]
            Returns :math:`x^2` where :math:`x` is the first column of `X`.
        """
        if not self.result_dot:
            raise Exception("Run the fit function first!")

        model = e.load_model(self.result_dot)

        return e.predict(X, model)

    def plot(self):
        """ Plot the state machine after fitting

        Returns
        -------
        A StateMachinePlot object implementing a _repr_svg_ method, used
        by Jupyter/IPython notebook to plot a graph
        """
        if has_graphviz:
            if self.result_dot is not None:
                with open("pre_"+self.result_dot) as result_dot_file:
                    return StateMachinePlot(
                       Source(result_dot_file.read())._repr_svg_()
                    )
            else:
                print("Run fit first!")
        else:
            print("This feature requires the graphviz python library")



import numpy as np
from numpy.testing import assert_almost_equal

from skltemplate import TemplateEstimator


def test_demo():
    X = np.random.random((100, 10))
    estimator = TemplateEstimator()
    estimator.fit(X, X[:, 0])
    assert_almost_equal(estimator.predict(X), X[:, 0]**2)


from __future__ import print_function
import sys
from setuptools import setup, find_packages

from distutils.command.build import build
from subprocess import call

import os

with open('requirements.txt') as f:
    INSTALL_REQUIRES = [l.strip() for l in f.readlines() if l]
with open('dependencies.txt') as f:
    DEPENDENCY_LINKS = [l.strip() for l in f.readlines() if l]
with open('setup.txt') as f:
    SETUP_REQUIRES = [l.strip() for l in f.readlines() if l]

BASEPATH = os.path.dirname(os.path.abspath(__file__))
SOURCE_DIR = 'src'

import sys, subprocess, pkg_resources

def missing_requirements(specifiers):
    for specifier in specifiers:
        try:
            pkg_resources.require(specifier)
        except pkg_resources.DistributionNotFound:
            yield specifier

def install_requirements_ez(specifiers):
    for to_install in list(specifiers):
        subprocess.call(["easy_install", to_install])

def install_requirements_pip(specifiers):
    for to_install in list(specifiers):
        subprocess.call([sys.executable, "-m", "pip", "install", to_install])

install_requirements_pip(missing_requirements(SETUP_REQUIRES))
install_requirements_ez(DEPENDENCY_LINKS)

class DFASATBuild(build):
    def run(self):
        # run original build code
        build.run(self)

        def exec_(cmd, cwd=None):
            call(cmd, cwd=cwd)

        # download DFASAT
        cmd = ['git',
                'clone',
                'https://bitbucket.org/chrshmmmr/dfasat.git',
                SOURCE_DIR
        ]

        self.execute(exec_, [cmd], 'Cloning dfasat')

        cmd = ['git', 'checkout', 'development'] # TODO: remove once python version is merged into master
        self.execute(exec_, [cmd, SOURCE_DIR], 'Checking out development')

        # build DFASAT
        build_path = os.path.abspath(self.build_temp)
        self.mkpath(build_path)

        cmd = [
            'make',
            'OUTDIR=' + build_path,
            'WITH_PYTHON=1'
        ]


        targets = ['clean', 'python']
        cmd.extend(targets)

        target_files = [os.path.join(build_path, file_) for file_ in ['dfasat.lib.so', 'dfasat.so']]

        def compile():
            call(cmd, cwd=os.path.join(BASEPATH, SOURCE_DIR))

        self.execute(compile, [], 'Compiling dfasat')

        lib_folder = os.path.join(self.build_lib, 'dfasat', 'lib')

        # copy resulting tool to library build folder
        self.mkpath(lib_folder)

        if not self.dry_run:
            for target in target_files:
                self.copy_file(target, lib_folder)

print(find_packages())

setup(name='dfasat',
      version='0.0.1.44',
      description='DFASAT',
      author='Christian Hammerschmidt, Benjamin Loos',
      packages=find_packages(),
#      dependency_links=DEPENDENCY_LINKS,
      install_requires=INSTALL_REQUIRES,
#      setup_requires=SETUP_REQUIRES,
      author_email='christian.hammerschmidt@uni.lu, loos.benjamin@gmail.com',
      package_dir = {'dfasat': 'dfasat'},
      zip_safe=False,
      #package_data = { 'dfasat': ["lib/dfasat.so", "lib/dfasat.lib.so"] },
      cmdclass = { 'build': DFASATBuild },
      )
